import inspect
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Type, get_type_hints

from pydantic import BaseConfig, BaseModel, create_model

from pait.exceptions import PaitBaseException
from pait.field import BaseField
from pait.g import config

from ._func_sig import FuncSig


class UndefinedType:
    def __repr__(self) -> str:
        return "PaitUndefined"


# pait undefined flag
Undefined: UndefinedType = UndefinedType()


def create_pydantic_model(
    annotation_dict: Dict[str, Tuple[Type, Any]],
    class_name: str = "DynamicModel",
    pydantic_config: Type[BaseConfig] = None,
    pydantic_base: Type[BaseModel] = None,
    pydantic_module: str = "pydantic.main",
    pydantic_validators: Dict[str, classmethod] = None,
) -> Type[BaseModel]:
    """pydantic create_model helper
    if use create_model('DynamicModel', **annotation_dict), mypy will tip error
    """
    return create_model(
        class_name,
        __config__=pydantic_config,
        __base__=pydantic_base,
        __module__=pydantic_module,
        __validators__=pydantic_validators,
        **annotation_dict,
    )


def gen_example_json_from_python(obj: Any) -> Any:
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = gen_example_json_from_python(value)
        return obj
    else:
        return config.python_type_default_value_dict.get(type(obj), obj)


def gen_example_dict_from_schema(
    schema_dict: Dict[str, Any], use_example_value: bool = False, definition_dict: Optional[dict] = None
) -> Dict[str, Any]:
    gen_dict: Dict[str, Any] = {}
    property_dict: Dict[str, Any] = schema_dict["properties"]
    if not definition_dict:
        _definition_dict: dict = schema_dict.get("definitions", {})
    else:
        _definition_dict = definition_dict
    for key, value in property_dict.items():
        if "items" in value and value["type"] == "array":
            if "$ref" in value["items"]:
                model_key: str = value["items"]["$ref"].split("/")[-1]
                gen_dict[key] = [
                    gen_example_dict_from_schema(
                        _definition_dict.get(model_key, {}), use_example_value, _definition_dict
                    )
                ]
            else:
                gen_dict[key] = []
        elif "$ref" in value:
            model_key = value["$ref"].split("/")[-1]
            gen_dict[key] = gen_example_dict_from_schema(
                _definition_dict.get(model_key, {}), use_example_value, _definition_dict
            )
        else:
            if "example" in value:
                gen_dict[key] = value["example"]
                if inspect.isfunction(gen_dict[key]):
                    gen_dict[key] = gen_dict[key]()
            elif "default" in value:
                gen_dict[key] = value["default"]
            else:
                if "type" in value:
                    if value["type"] not in config.json_type_default_value_dict:
                        raise KeyError(f"Can not found type: {key} in json type")
                    gen_dict[key] = config.json_type_default_value_dict[value["type"]]
                else:
                    gen_dict[key] = "object()"
    return gen_dict


def gen_example_json_from_schema(schema_dict: Dict[str, Any], use_example_value: bool) -> str:
    return json.dumps(
        gen_example_dict_from_schema(schema_dict, use_example_value=use_example_value), cls=config.json_encoder
    )


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: List["inspect.Parameter"] = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            default: Any = getattr(cbv_class, param_name, Undefined)
            if not isinstance(default, BaseField):
                continue
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=default,
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    return parameter_list


def gen_tip_exc(_object: Any, exception: "Exception", parameter: Optional[inspect.Parameter] = None) -> Exception:
    """Help users understand which parameter is wrong"""
    if getattr(exception, "_is_tip_exc", None):
        return exception
    if parameter:
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        parameter_value_name: str = param_value.__class__.__name__
        if param_value is parameter.empty:
            param_str: str = f"{param_name}: {annotation}"
        else:
            param_str = f"{param_name}: {annotation} = {parameter_value_name}"
            if isinstance(param_value, BaseField):
                param_str += f"(alias={param_value.alias}, default={param_value.default})"
    else:
        param_str = ""

    file: Optional[str] = None
    if isinstance(_object, FuncSig):
        _object = _object.func
    if inspect.isfunction(_object):
        title: str = "def"
        if inspect.iscoroutinefunction(_object):
            title = "async def"
        file = inspect.getfile(_object)
        line: int = inspect.getsourcelines(_object)[1]
        error_object_name: str = _object.__name__
        logging.debug(
            f"""
{title} {error_object_name}(
    ...
    {param_str} <-- error
    ...
):
    pass
"""
        )
    else:
        module: Any = inspect.getmodule(_object)
        if module:
            file = module.__file__
        line = inspect.getsourcelines(_object.__class__)[1]
        error_object_name = _object.__class__.__name__
        if "class" in error_object_name:
            error_object_name = str(_object.__class__)
        logging.debug(f"class: `{error_object_name}`  attributes error\n    {param_str}")
    exc_msg: str = f'File "{file}",' f" line {line}," f" in {error_object_name}." f" error:{str(exception)}"
    try:
        exc: Exception = exception.__class__(exc_msg)
    except Exception:
        exc = PaitBaseException(exc_msg)
    setattr(exception, "_is_tip_exc", True)
    return exc



from typing import Optional, Tuple, Type

from pydantic import BaseModel


class PaitResponseModel(object):
    """response model"""

    # response name
    name: Optional[str] = None
    # response description
    description: Optional[str] = None
    # response header
    header: dict = {}
    # response media type
    media_type: str = "application/json"
    # response data
    response_data: Optional[Type[BaseModel]] = None
    # response status code
    status_code: Tuple[int] = (200,)



import asyncio
import inspect
import logging
from typing import Any, Callable, Coroutine, Dict, List, Mapping, Optional, Tuple, Type, Union, get_type_hints

from pydantic import BaseModel, fields

from pait import field
from pait.app.base import BaseAppHelper
from pait.exceptions import NotFoundFieldError, PaitBaseException
from pait.field import BaseField
from pait.model.base_model import PaitBaseModel
from pait.util import FuncSig, create_pydantic_model, gen_tip_exc, get_func_sig, get_parameter_list_from_class


def parameter_2_basemodel(parameter_value_dict: Dict["inspect.Parameter", Any]) -> BaseModel:
    """Convert all parameters into pydantic mods"""
    annotation_dict: Dict[str, Tuple[Type, Any]] = {}
    param_value_dict: Dict[str, Any] = {}
    for parameter, value in parameter_value_dict.items():
        annotation_dict[parameter.name] = (parameter.annotation, parameter.default)
        key: str = parameter.name
        if isinstance(parameter.default, BaseField) and parameter.default.alias:
            key = parameter.default.alias
        param_value_dict[key] = value

    dynamic_model: Type[BaseModel] = create_pydantic_model(annotation_dict)
    return dynamic_model(**param_value_dict)


def get_request_value_from_parameter(
    parameter: inspect.Parameter, app_helper: "BaseAppHelper"
) -> Union[Any, Coroutine]:
    field_name: str = parameter.default.__class__.__name__.lower()
    # Note: not use hasattr with LazyProperty (
    #   because hasattr will calling getattr(obj, name) and catching AttributeError,
    # )
    app_field_func: Optional[Callable] = getattr(app_helper, field_name, None)
    if app_field_func is None:
        raise NotFoundFieldError(f"field: {field_name} not found in {app_helper}")
    return app_field_func()


def _set_parameter_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> bool:
    """use func_args param faster return and extend func_args"""
    if parameter.name == "self" and not func_args:
        # Only support self param name
        func_args.append(app_helper.cbv_class)
    elif app_helper.check_request_type(parameter.annotation):
        # support request param(def handle(request: Request))
        func_args.append(app_helper.request)
    elif issubclass(parameter.annotation, PaitBaseModel):
        return True
    else:
        logging.warning(f"Pait not support args: {parameter}")
    return False


def set_parameter_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> None:
    """use func_args param faster return and extend func_args"""
    if _set_parameter_value_to_args(parameter, app_helper, func_args):
        # support pait_model param(def handle(model: PaitBaseModel))
        single_field_dict: Dict["inspect.Parameter", Any] = {}
        _pait_model: Type[PaitBaseModel] = parameter.annotation
        for param_name, param_annotation in get_type_hints(_pait_model).items():
            if param_name.startswith("_"):
                continue
            parameter = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=getattr(_pait_model, param_name),
                annotation=param_annotation,
            )
            request_value: Any = get_request_value_from_parameter(parameter, app_helper)
            # PaitModel's attributes's annotation support BaseModel
            request_value_handle(parameter, request_value, None, single_field_dict, app_helper)
        func_args.append(parameter_2_basemodel(single_field_dict))


async def async_set_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> None:
    """use func_args param faster return and extend func_args"""
    if _set_parameter_value_to_args(parameter, app_helper, func_args):
        # support pait_model param(def handle(model: PaitModel))
        single_field_dict: Dict["inspect.Parameter", Any] = {}
        _pait_model: Type[PaitBaseModel] = parameter.annotation

        for param_name, param_annotation in get_type_hints(_pait_model).items():
            if param_name.startswith("_"):
                continue
            parameter = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=getattr(_pait_model, param_name),
                annotation=param_annotation,
            )
            request_value: Any = get_request_value_from_parameter(parameter, app_helper)
            if asyncio.iscoroutine(request_value):
                request_value = await request_value
            # PaitModel's attributes  support BaseModel
            request_value_handle(parameter, request_value, None, single_field_dict, app_helper)
        func_args.append(parameter_2_basemodel(single_field_dict))


def request_value_handle(
    parameter: inspect.Parameter,
    request_value: Any,
    base_model_dict: Optional[Dict[str, Any]],
    parameter_value_dict: Dict["inspect.Parameter", Any],
    app_helper: "BaseAppHelper",
) -> None:
    """parse request_value and set to base_model_dict or parameter_value_dict"""
    param_value: BaseField = parameter.default
    annotation: Type[BaseModel] = parameter.annotation
    param_name: str = parameter.name

    if not isinstance(param_value, BaseField):
        raise PaitBaseException(f"must use {BaseField.__class__.__name__}, no {param_value}")
    if param_value.raw_return:
        parameter_value_dict[parameter] = request_value
    elif (
        isinstance(request_value, Mapping)
        or app_helper.check_header_type(type(request_value))
        or app_helper.check_form_type(type(request_value))
    ):

        if base_model_dict is not None and inspect.isclass(annotation) and issubclass(annotation, BaseModel):
            # parse annotation is pydantic.BaseModel and base_model_dict not None
            base_model_dict[parameter.name] = annotation(**request_value)
        else:
            # parse annotation is python type and pydantic.field
            if type(param_value.alias) is str and param_value.alias in request_value:
                value = request_value[param_value.alias]
            elif param_name in request_value:
                value = request_value[param_name]
            else:
                # not found value
                if param_value.default != fields.Undefined:
                    value = param_value.default
                else:
                    value = fields.Undefined
            parameter_value_dict[parameter] = value
    else:
        parameter_value_dict[parameter] = request_value


def param_handle(
    app_helper: "BaseAppHelper", _object: Union[FuncSig, Type], param_list: List["inspect.Parameter"]
) -> Tuple[List[Any], Dict[str, Any]]:
    args_param_list: List[Any] = []
    kwargs_param_dict: Dict[str, Any] = {}
    single_field_dict: Dict["inspect.Parameter", Any] = {}

    for parameter in param_list:
        try:
            if parameter.default != parameter.empty:
                # kwargs param
                # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                if isinstance(parameter.default, field.Depends):
                    func: Callable = parameter.default.func
                    func_sig: FuncSig = get_func_sig(func)
                    _func_args, _func_kwargs = param_handle(app_helper, func_sig, func_sig.param_list)
                    func_result: Any = func(*_func_args, **_func_kwargs)
                    kwargs_param_dict[parameter.name] = func_result
                else:
                    request_value: Any = get_request_value_from_parameter(parameter, app_helper)
                    request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict, app_helper)
            else:
                # args param
                # support model: model: ModelType
                set_parameter_value_to_args(parameter, app_helper, args_param_list)
        except PaitBaseException as e:
            raise e from gen_tip_exc(_object, e, parameter)
    # support field: def demo(demo_param: int = pait.field.BaseField())
    if single_field_dict:
        try:
            kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).dict())
        except Exception as e:
            raise e from gen_tip_exc(_object, e)
    return args_param_list, kwargs_param_dict


async def async_param_handle(
    dispatch_web: "BaseAppHelper", _object: Union[FuncSig, Type], param_list: List["inspect.Parameter"]
) -> Tuple[List[Any], Dict[str, Any]]:
    args_param_list: List[Any] = []
    kwargs_param_dict: Dict[str, Any] = {}
    single_field_dict: Dict["inspect.Parameter", Any] = {}
    for parameter in param_list:
        try:
            if parameter.default != parameter.empty:
                # kwargs param
                # support model: pydantic.BaseModel = pait.field.BaseField()
                if isinstance(parameter.default, field.Depends):
                    func: Callable = parameter.default.func
                    func_sig: FuncSig = get_func_sig(func)
                    _func_args, _func_kwargs = await async_param_handle(dispatch_web, func_sig, func_sig.param_list)
                    func_result: Any = func(*_func_args, **_func_kwargs)
                    if asyncio.iscoroutine(func_result):
                        func_result = await func_result
                    kwargs_param_dict[parameter.name] = func_result
                else:
                    request_value: Any = get_request_value_from_parameter(parameter, dispatch_web)

                    if asyncio.iscoroutine(request_value) or asyncio.isfuture(request_value):
                        request_value = await request_value

                    request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict, dispatch_web)
            else:
                # args param
                # support model: model: ModelType
                await async_set_value_to_args(parameter, dispatch_web, args_param_list)
        except PaitBaseException as e:
            raise e from gen_tip_exc(_object, e, parameter)

    # support field: def demo(demo_param: int = pait.field.BaseField())
    if single_field_dict:
        try:
            try:
                kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).dict())
            except TypeError:
                kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).__dict__)

        except Exception as e:
            raise e from gen_tip_exc(_object, e)
    return args_param_list, kwargs_param_dict


async def async_class_param_handle(dispatch_app: "BaseAppHelper") -> None:
    cbv_class: Optional[Type] = dispatch_app.cbv_class
    if not cbv_class:
        return
    param_list: list = get_parameter_list_from_class(cbv_class)
    args, kwargs = await async_param_handle(dispatch_app, cbv_class, param_list)
    cbv_class.__dict__.update(kwargs)


def class_param_handle(dispatch_app: "BaseAppHelper") -> None:
    cbv_class: Optional[Type] = dispatch_app.cbv_class
    if not cbv_class:
        return
    param_list: list = get_parameter_list_from_class(cbv_class)
    args, kwargs = param_handle(dispatch_app, cbv_class, param_list)
    cbv_class.__dict__.update(kwargs)


async def async_func_param_handle(dispatch_app: "BaseAppHelper", func_sig: FuncSig) -> Tuple[List[Any], Dict[str, Any]]:
    return await async_param_handle(dispatch_app, func_sig, func_sig.param_list)


def func_param_handle(dispatch_app: "BaseAppHelper", func_sig: FuncSig) -> Tuple[List[Any], Dict[str, Any]]:
    return param_handle(dispatch_app, func_sig, func_sig.param_list)



import inspect
from typing import Callable, Dict, Type, Union

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.routing import Route

from pait.g import pait_data
from pait.model.core import PaitCoreModel

from ._app_helper import AppHelper

__all__ = ["load_app"]


def load_app(app: Starlette, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for route in app.routes:
        if not isinstance(route, Route):
            # not support
            continue
        path: str = route.path
        method_set: set = route.methods or set()
        route_name: str = route.name
        endpoint: Union[Callable, Type] = route.endpoint
        pait_id: str = getattr(route.endpoint, "_pait_id", None)
        if not pait_id and inspect.isclass(endpoint) and issubclass(endpoint, HTTPEndpoint):  # type: ignore
            for method in ["get", "post", "head", "options", "delete", "put", "trace", "patch"]:
                method_endpoint = getattr(endpoint, method, None)
                if not method_endpoint:
                    continue
                method_set = {method}
                pait_id = getattr(method_endpoint, "_pait_id", None)
                if not pait_id:
                    continue
                pait_data.add_route_info(
                    AppHelper.app_name, pait_id, path, method_set, f"{route_name}.{method}", project_name
                )
                _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
        elif pait_id:
            pait_data.add_route_info(AppHelper.app_name, pait_id, path, method_set, route_name, project_name)
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data



from enum import Enum


class PaitStatus(Enum):
    """Interface life cycle"""

    undefined = "undefined"
    # The interface is under development and will frequently change
    design = "design"
    dev = "dev"

    # The interface has been completed, but there may be some bugs
    integration = "integration"
    complete = "complete"
    test = "test"

    # The interface is online
    pre_release = "pre_release"
    release = "release"

    # The interface has been online, but needs to be offline for some reasons
    abnormal = "abnormal"
    maintenance = "maintenance"
    archive = "archive"
    abandoned = "abandoned"



from typing import Callable, Dict, Optional

from tornado.web import Application

from pait.g import pait_data
from pait.model.core import PaitCoreModel

from ._app_helper import AppHelper

__all__ = ["load_app"]


def load_app(app: Application, project_name: str = "") -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for rule in app.wildcard_router.rules:
        path: str = rule.matcher.regex.pattern  # type: ignore
        base_name: str = rule.target.__name__
        for method in ["get", "post", "head", "options", "delete", "put", "trace", "patch"]:
            try:
                handler: Callable = getattr(rule.target, method, None)
            except TypeError:
                continue
            route_name: str = f"{base_name}.{method}"
            pait_id: Optional[str] = getattr(handler, "_pait_id", None)
            if not pait_id:
                continue
            pait_data.add_route_info(AppHelper.app_name, pait_id, path, {method}, route_name, project_name)
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data



import json
from typing import Any, Dict, List, Mapping, Tuple

from tornado.httputil import RequestStartLine

from pait.app.base import BaseAppHelper
from pait.util import LazyProperty

__all__ = ["AppHelper"]


class AppHelper(BaseAppHelper):
    RequestType = RequestStartLine
    FormType = dict
    FileType = dict
    HeaderType = dict
    app_name = "tornado"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)
        if not self.cbv_class:
            raise RuntimeError("Can not load Tornado handle")

        self.request = self.cbv_class.request
        self.path_kwargs: Dict[str, Any] = self.cbv_class.path_kwargs

    def body(self) -> dict:
        return json.loads(self.request.body.decode())

    def cookie(self) -> dict:
        return self.request.cookies

    def file(self) -> dict:
        return {item["filename"]: item for item in self.request.files["file"]}

    @LazyProperty(is_class_func=True)
    def form(self) -> dict:
        if self.request.arguments:
            form_dict: dict = {key: value[0].decode() for key, value in self.request.arguments.items()}
        else:
            form_dict = json.loads(self.request.body.decode())
        return {key: value[0] for key, value in form_dict.items()}

    def header(self) -> dict:
        return self.request.headers

    def path(self) -> dict:
        return self.path_kwargs

    @LazyProperty(is_class_func=True)
    def query(self) -> dict:
        return {key: value[0].decode() for key, value in self.request.query_arguments.items()}

    @LazyProperty(is_class_func=True)
    def multiform(self) -> Dict[str, List[Any]]:
        if self.request.arguments:
            return {key: [i.decode() for i in value] for key, value in self.request.arguments.items()}
        else:
            return {key: [value] for key, value in json.loads(self.request.body.decode()).items()}

    @LazyProperty(is_class_func=True)
    def multiquery(self) -> Dict[str, Any]:
        return {key: [i.decode() for i in value] for key, value in self.request.query_arguments.items()}



import logging
from typing import Any, Dict, List, Optional

from sanic import response
from sanic.app import Sanic
from sanic.blueprints import Blueprint
from sanic.exceptions import NotFound
from sanic.request import Request
from sanic.response import HTTPResponse
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.api_doc.html import get_redoc_html as _get_redoc_html
from pait.api_doc.html import get_swagger_ui_html as _get_swagger_ui_html
from pait.api_doc.open_api import PaitOpenApi
from pait.model.core import PaitCoreModel

from ._load_app import load_app

__all__ = ["add_doc_route"]


def add_doc_route(
    app: Sanic,
    prefix: str = "/",
    pin_code: str = "",
    title: str = "Pait Doc",
    open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
) -> None:
    if pin_code:
        logging.info(f"doc route start pin code:{pin_code}")

    def _get_request_pin_code(request: Request) -> Optional[str]:
        r_pin_code: Optional[str] = request.args.get("pin_code", None)
        if pin_code:
            if r_pin_code != pin_code:
                raise NotFound
        return r_pin_code

    def _get_open_json_url(request: Request) -> str:
        r_pin_code: Optional[str] = _get_request_pin_code(request)
        openapi_json_url: str = f"http://{request.host}{'/'.join(request.path.split('/')[:-1])}/openapi.json"
        if r_pin_code:
            openapi_json_url += f"?pin_code={r_pin_code}"
        return openapi_json_url

    def get_redoc_html(request: Request) -> HTTPResponse:
        return response.html(_get_redoc_html(_get_open_json_url(request), title))

    def get_swagger_ui_html(request: Request) -> HTTPResponse:
        return response.html(_get_swagger_ui_html(_get_open_json_url(request), title))

    def openapi_route(request: Request) -> HTTPResponse:
        _get_request_pin_code(request)
        pait_dict: Dict[str, PaitCoreModel] = load_app(request.app)
        pait_openapi: PaitOpenApi = PaitOpenApi(
            pait_dict,
            title=title,
            open_api_server_list=[{"url": f"http://{request.host}", "description": ""}],
            open_api_tag_list=open_api_tag_list,
        )
        return response.json(pait_openapi.open_api_dict)

    blueprint: Blueprint = Blueprint("api doc", prefix)
    blueprint.add_route(get_redoc_html, "/redoc", methods={"GET"})
    blueprint.add_route(get_swagger_ui_html, "/swagger", methods={"GET"})
    blueprint.add_route(openapi_route, "/openapi.json", methods={"GET"})
    app.blueprint(blueprint)



from typing import Any, Coroutine, Dict, List, Mapping, Optional, Tuple

from starlette.datastructures import FormData, Headers, UploadFile
from starlette.requests import Request

from pait.app.base import BaseAppHelper
from pait.util import LazyProperty

__all__ = ["AppHelper"]


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = FormData
    FileType = UploadFile
    HeaderType = Headers
    app_name = "starlette"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)
        self._form: Optional[FormData] = None

    def body(self) -> dict:
        return self.request.json()

    def cookie(self) -> dict:
        return self.request.cookies

    async def get_form(self) -> FormData:
        if self._form:
            return self._form
        form: FormData = await self.request.form()
        self._form = form
        return form

    def file(self) -> Coroutine[Any, Any, FormData]:
        async def _() -> FormData:
            return await self.get_form()

        return _()

    def form(self) -> Coroutine[Any, Any, Dict[str, Any]]:
        @LazyProperty()
        async def _form() -> Dict[str, Any]:
            form_data: FormData = await self.get_form()
            return {key: form_data.getlist(key)[0] for key, _ in form_data.items()}

        return _form()

    def header(self) -> Headers:
        return self.request.headers

    def path(self) -> dict:
        return self.request.path_params

    def query(self) -> dict:
        return dict(self.request.query_params)

    def multiform(self) -> Coroutine[Any, Any, Dict[str, List[Any]]]:
        @LazyProperty()
        async def _multiform() -> Dict[str, List[Any]]:
            form_data: FormData = await self.get_form()
            return {
                key: [i for i in form_data.getlist(key) if not isinstance(i, UploadFile)]
                for key, _ in form_data.items()
            }

        return _multiform()

    @LazyProperty(is_class_func=True)
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.query_params.getlist(key) for key, _ in self.request.query_params.items()}



import argparse
import importlib
import inspect
from typing import Dict, List, Optional

from pait.api_doc.markdown import PaitMd
from pait.api_doc.open_api import PaitOpenApi
from pait.app import load_app
from pait.model.core import PaitCoreModel


def main() -> None:
    output_type_list: List[str] = ["md", "openapi_json", "openapi_yaml"]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "app",
        help="Enter the file name and app name of the app that needs to generate the document,"
        " for example:app stands for. /example.py app",
    )
    parser.add_argument("-o", "--output_type", help=f"now support: {','.join(output_type_list)}")
    parser.add_argument("-t", "--title", help="Title of doc", default="Pait Doc")
    parser.add_argument("-f", "--filename", help="Name of save file", default=None)
    parser.add_argument(
        "-u", "--use_html_details", help="Whether to output html text in the md document", default=True, type=bool
    )
    parser.add_argument("-i", "--indent", help="json formatted output of indent values", default=2, type=int)

    args, unknown = parser.parse_known_args()

    module_name, app_name = args.app.split(":")
    input_output_type_list: List[str] = [i for i in args.output_type.split(",") if i in output_type_list]
    title: str = args.title
    filename: Optional[str] = args.filename
    use_html_details: bool = args.use_html_details

    module = importlib.import_module(module_name)
    app = getattr(module, app_name, None)
    if not app:
        raise ImportError(f"Can't found {app} in {module}")
    if inspect.isfunction(app):
        app = app()
    pait_dict: Dict[str, PaitCoreModel] = load_app(app)

    for type_ in input_output_type_list:
        if type_ == "md":
            PaitMd(pait_dict, title=title, use_html_details=use_html_details).output(filename)
        elif type_ == "openapi_json":
            PaitOpenApi(pait_dict, title=title, type_="json").output(filename)
        elif type_ == "openapi_yaml":
            PaitOpenApi(pait_dict, title=title, type_="yaml").output(filename)
        else:
            raise ValueError(f"Not support output type: {type_}, please select from the following options")


if __name__ == "__main__":
    main()



import asyncio
import inspect
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Set, Tuple, Type

from pait.g import config
from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus

if TYPE_CHECKING:
    from pait.app.base import BaseAppHelper


class PaitCoreModel(object):
    def __init__(
        self,
        func: Callable,
        app_helper_class: "Type[BaseAppHelper]",
        make_mock_response_fn: Callable[[Type[PaitResponseModel]], Any],
        path: Optional[str] = None,
        method_set: Optional[Set[str]] = None,
        operation_id: Optional[str] = None,
        func_name: Optional[str] = None,
        author: Optional[Tuple[str, ...]] = None,
        summary: Optional[str] = None,
        desc: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
    ):
        self.app_helper_class: "Type[BaseAppHelper]" = app_helper_class
        self.make_mock_response_fn: Callable[[Type[PaitResponseModel]], Any] = make_mock_response_fn
        self._func: Callable = func  # route func
        self.qualname: str = func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        self.pait_id: str = f"{self.qualname}_{id(func)}"
        setattr(func, "_pait_id", self.pait_id)
        self.path: str = path or ""  # request url path
        self._method_list: List[str] = sorted(list(method_set or set()))  # request method set
        self.func_name: str = func_name or func.__name__
        self.operation_id: str = operation_id or self.func_name  # route name
        self._author: Optional[Tuple[str, ...]] = author  # The main developer of this func
        self.summary: str = summary or ""
        self.desc: str = desc or func.__doc__ or ""  # desc of this func
        self._status: Optional[PaitStatus] = status  # Interface development progress (life cycle)
        self.group: str = group or "root"  # Which group this interface belongs to
        self.tag: Tuple[str, ...] = tag or ("default",)  # Interface tag
        self._response_model_list: List[Type[PaitResponseModel]] = response_model_list or []
        self.func_path: str = ""

    @property
    def author(self) -> Tuple[str, ...]:
        return self._author or config.author

    @property
    def status(self) -> PaitStatus:
        return self._status or config.status

    @property
    def func(self) -> Callable:
        if config.enable_mock_response:
            return self.return_mock_response
        return self._func

    @property
    def method_list(self) -> List[str]:
        return self._method_list

    @method_list.setter
    def method_list(self, method_list: List[str]) -> None:
        _temp_set: Set[str] = set(self._method_list) | set(method_list)
        _temp_set.difference_update(config.block_http_method_set)
        self._method_list = sorted(list(_temp_set))

    def return_mock_response(self, *args: Any, **kwargs: Any) -> Any:
        if not self.response_model_list:
            raise RuntimeError(f"{self.func} can not set response model")
        pait_response: Optional[Type[PaitResponseModel]] = None
        if config.enable_mock_response_filter_fn:
            for _pait_response in self.response_model_list:
                if config.enable_mock_response_filter_fn(_pait_response):
                    pait_response = _pait_response
                    break
        if not pait_response:
            pait_response = self.response_model_list[0]

        # fix tornado
        if self.app_helper_class.app_name == "tornado":
            setattr(pait_response, "handle", args[0])
        resp: Any = self.make_mock_response_fn(pait_response)
        # support async def
        if inspect.iscoroutinefunction(self._func):
            future: asyncio.Future = asyncio.Future()
            future.set_result(resp)
            resp = future
        return resp

    @property
    def response_model_list(self) -> List[Type[PaitResponseModel]]:
        if config.default_response_model_list:
            return self._response_model_list + config.default_response_model_list
        else:
            return self._response_model_list



import inspect
import sys
from dataclasses import dataclass
from typing import Callable, ForwardRef, List


@dataclass()
class FuncSig:
    """func inspect.Signature model"""

    func: Callable
    sig: "inspect.Signature"
    param_list: List["inspect.Parameter"]


def get_func_sig(func: Callable) -> FuncSig:
    """get func inspect.Signature model"""
    sig: inspect.Signature = inspect.signature(func)
    param_list: List[inspect.Parameter] = []
    for key in sig.parameters:
        if not (sig.parameters[key].annotation != sig.empty or sig.parameters[key].name == "self"):
            continue
        parameter: inspect.Parameter = sig.parameters[key]
        if isinstance(parameter.annotation, str):
            value: ForwardRef = ForwardRef(parameter.annotation, is_argument=False)
            setattr(
                parameter, "_annotation", value._evaluate(sys.modules[func.__module__].__dict__, None)  # type: ignore
            )
        param_list.append(parameter)

    # return_param = sig.return_annotation
    return FuncSig(func=func, sig=sig, param_list=param_list)



import copy
import difflib
import inspect
import logging
from typing import Any, Callable, Dict, Generic, List, Mapping, Optional, Tuple, Type, TypeVar
from urllib.parse import urlencode

from pydantic import BaseModel, ValidationError

from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel
from pait.util import gen_example_dict_from_schema, gen_example_json_from_python


class BaseAppHelper(object):
    """Provide a unified framework call interface for pait"""

    RequestType = type(None)
    FormType = type(None)
    FileType = type(None)
    HeaderType = type(None)
    app_name: str = "BaseAppHelper"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        """
        Extract the required data from the passed parameters,
        such as the self parameter in cvb mode, the request parameter in starletter
        """
        self.cbv_class: Any = None

        request = None
        new_args: List[Any] = []
        for param in args:
            if type(param) == self.RequestType:
                request = param
                # In cbv, request parameter will only appear after the self parameter
                break
            elif isinstance(param, class_):
                self.cbv_class = param
            else:
                # In cbv, parameter like self, request, {other param}
                # Now, not support other param
                logging.warning("Pait only support self and request args param")
                break
            new_args.append(param)

        self.request: Any = request
        self.request_args: List[Any] = new_args
        self.request_kwargs: Mapping[str, Any] = kwargs

    def cookie(self) -> Any:
        raise NotImplementedError

    def header(self) -> Any:
        raise NotImplementedError

    def path(self) -> Any:
        raise NotImplementedError

    def query(self) -> Any:
        raise NotImplementedError

    def body(self) -> Any:
        raise NotImplementedError

    def file(self) -> Any:
        raise NotImplementedError

    def form(self) -> Any:
        raise NotImplementedError

    def multiform(self) -> Any:
        raise NotImplementedError

    def multiquery(self) -> Any:
        raise NotImplementedError

    def check_request_type(self, value: Any) -> bool:
        return value is self.RequestType

    def check_file_type(self, value: Any) -> bool:
        return value is self.FileType

    def check_form_type(self, value: Any) -> bool:
        return value is self.FormType

    def check_header_type(self, value: Any) -> bool:
        return value is self.HeaderType


RESP_T = TypeVar("RESP_T")


class BaseTestHelper(Generic[RESP_T]):
    client: Any

    def __init__(
        self,
        client: Any,
        func: Callable,
        pait_dict: Optional[Dict[str, PaitCoreModel]] = None,
        body_dict: Optional[dict] = None,
        cookie_dict: Optional[dict] = None,
        file_dict: Optional[dict] = None,
        form_dict: Optional[dict] = None,
        header_dict: Optional[dict] = None,
        path_dict: Optional[dict] = None,
        query_dict: Optional[dict] = None,
    ):
        """
        :param client:  test client
        :param func:  route handle decorated by pait
        :param pait_dict:  pait core data
        :param body_dict:  request body param
        :param cookie_dict:  request cookie param
        :param file_dict:  request file param
        :param form_dict:  request form param
        :param header_dict:  request herder param
        :param path_dict:  request path param
        :param query_dict:  request query param
        """
        pait_id: str = getattr(func, "_pait_id", "")
        if not pait_id:
            raise RuntimeError(f"Can not found pait id from {func}")

        self.client: Any = client
        self.func: Callable = func
        # pait dict handle
        if pait_dict:
            self.pait_dict: Dict[str, PaitCoreModel] = pait_dict
        else:
            self.pait_dict = self._gen_pait_dict()
        self.pait_core_model: PaitCoreModel = self.pait_dict[pait_id]

        self.body_dict: Optional[dict] = body_dict
        self.cookie_dict: Optional[dict] = cookie_dict
        self.file_dict: Optional[dict] = file_dict
        self.form_dict: Optional[dict] = form_dict
        self.header_dict: dict = header_dict or {}
        self.path_dict: Optional[dict] = path_dict
        self.query_dict: Optional[dict] = query_dict

        # path handle
        self.path: str = self.pait_core_model.path
        if self.path_dict:
            path_list: List[str] = self.path.split("/")
            new_path_list: List[str] = []
            for sub_path in path_list:
                if not sub_path:
                    continue
                new_sub_path: Optional[str] = self._replace_path(sub_path)
                if new_sub_path:
                    new_path_list.append(new_sub_path)
                else:
                    new_path_list.append(sub_path)
            self.path = "/".join([str(i) for i in new_path_list])

        # add query path by path
        if self.query_dict:
            self.path = self.path + "?" + urlencode(self.query_dict, True)
        if not self.path.startswith("/"):
            self.path = "/" + self.path

        # auto select method
        self.method: Optional[str] = None
        if len(self.pait_core_model.method_list) == 1:
            self.method = self.pait_core_model.method_list[0]
        elif "GET" in self.pait_core_model.method_list:
            self.method = "GET"
        elif "POST" in self.pait_core_model.method_list:
            self.method = "POST"

        self._app_init_field()

    def _app_init_field(self) -> None:
        """init request param by application framework"""
        raise NotImplementedError()

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        """load pait dict"""
        raise NotImplementedError()

    @staticmethod
    def _check_resp_status(resp: RESP_T, response_model: Type[PaitResponseModel]) -> bool:
        raise NotImplementedError()

    @staticmethod
    def _check_resp_media_type(resp: RESP_T, response_model: Type[PaitResponseModel]) -> bool:
        raise NotImplementedError()

    @staticmethod
    def _get_json(resp: RESP_T) -> dict:
        raise NotImplementedError()

    def _diff_resp_dict(self, raw_resp: Any, default_resp: Any, parent_key_list: Optional[List[str]] = None) -> bool:
        """check resp data structure diff"""
        raw_parent_key_list: List[str] = parent_key_list or []
        try:
            if isinstance(raw_resp, dict):
                for key, value in raw_resp.items():
                    parent_key_list = copy.deepcopy(raw_parent_key_list)
                    parent_key_list.append(key)
                    if not self._diff_resp_dict(value, default_resp[key], parent_key_list):
                        return False
                return True
            elif isinstance(raw_resp, list) or isinstance(raw_resp, tuple):
                if (
                    raw_resp
                    and default_resp
                    and not self._diff_resp_dict(raw_resp[0], default_resp[0], parent_key_list)
                ):
                    return False
                return True
            else:
                return True
        except KeyError:
            raise RuntimeError(f"Can not found key from model, key:{' -> '.join(parent_key_list or '')}")

    def _assert_response(self, resp: RESP_T) -> None:
        """Whether the structure of the check response is correct"""
        if not self.pait_core_model.response_model_list:
            return

        json_error_response_dict: Dict[Type[BaseModel], Tuple[Exception, float]] = {}
        for response_model in self.pait_core_model.response_model_list:
            check_list: List[bool] = [
                self._check_resp_status(resp, response_model),
                self._check_resp_media_type(resp, response_model),
            ]
            if response_model.media_type == "application/json":
                response_data_model: Any = response_model.response_data
                if response_data_model:
                    if not (inspect.isclass(response_data_model) and issubclass(response_data_model, BaseModel)):
                        raise TypeError(f"{response_data_model} not issubclass {BaseModel}")
                    resp_dict: Optional[dict] = self._get_json(resp)
                    if not resp_dict:
                        check_list.append(True)
                        continue
                    response_data_default_dict: dict = gen_example_dict_from_schema(response_data_model.schema())
                    try:
                        response_data_model(**resp_dict)
                        check_list.append(self._diff_resp_dict(resp_dict, response_data_default_dict))
                    except (ValidationError, RuntimeError) as e:
                        check_list.append(False)
                        json_error_response_dict[response_data_model] = (
                            e,
                            difflib.SequenceMatcher(
                                None,
                                str(gen_example_json_from_python(resp_dict.copy())),
                                str(response_data_default_dict),
                            ).quick_ratio(),
                        )
            if all(check_list):
                return
        if json_error_response_dict:
            exc: Optional[Exception] = None
            model: Optional[Type[BaseModel]] = None
            max_quick_ratio: float = 0.0
            for _model, response_error in json_error_response_dict.items():
                if response_error[1] > max_quick_ratio:
                    max_quick_ratio = response_error[1]
                    exc = response_error[0]
                    model = _model
            if exc:
                raise RuntimeError(f"maybe error:{exc} by response_model: {model}") from exc
        raise RuntimeError(f"response check error by:{self.pait_core_model.response_model_list}.")

    def _replace_path(self, path_str: str) -> Optional[str]:
        raise NotImplementedError()

    def _make_response(self, method: str) -> RESP_T:
        """Whether the structure of the check response is correct"""
        raise NotImplementedError()

    def request(self, method: Optional[str] = None) -> RESP_T:
        """user call test request api"""
        if not method:
            method = self.method
        if not method:
            raise ValueError("Method is Null")
        resp: RESP_T = self._make_response(method)
        if self.pait_core_model.response_model_list:
            self._assert_response(resp)
        return resp

    def get(self) -> RESP_T:
        return self.request("GET")

    def patch(self) -> RESP_T:
        return self.request("PATCH")

    def post(self) -> RESP_T:
        return self.request("POST")

    def head(self) -> RESP_T:
        return self.request("HEAD")

    def put(self) -> RESP_T:
        return self.request("PUT")

    def delete(self) -> RESP_T:
        return self.request("DELETE")

    def options(self) -> RESP_T:
        return self.request("OPTIONS")



from typing import Dict, Optional, Type

from requests import Response as _Response  # type: ignore
from starlette.testclient import TestClient

from pait.app.base import BaseTestHelper
from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel

from ._load_app import load_app

__all__ = ["StarletteTestHelper"]


class StarletteTestHelper(BaseTestHelper[_Response]):
    client: TestClient

    def _app_init_field(self) -> None:
        if self.cookie_dict:
            self.header_dict.update(self.cookie_dict)

    def _gen_pait_dict(self) -> Dict[str, PaitCoreModel]:
        return load_app(self.client.app)  # type: ignore

    @staticmethod
    def _check_resp_status(resp: _Response, response_model: Type[PaitResponseModel]) -> bool:
        return resp.status_code in response_model.status_code

    @staticmethod
    def _check_resp_media_type(resp: _Response, response_model: Type[PaitResponseModel]) -> bool:
        return resp.headers["content-type"] == response_model.media_type

    @staticmethod
    def _get_json(resp: _Response) -> dict:
        return resp.json()

    def _replace_path(self, path_str: str) -> Optional[str]:
        if self.path_dict and path_str[0] == "{" and path_str[-1] == "}":
            return self.path_dict[path_str[1:-1]]
        return None

    def _make_response(self, method: str) -> _Response:
        method = method.upper()
        resp: _Response = self.client.request(
            method,
            url=self.path,
            cookies=self.cookie_dict,
            data=self.form_dict,
            json=self.body_dict,
            headers=self.header_dict,
            files=self.file_dict,
        )
        return resp



import inspect
import json
from types import CodeType
from typing import Any, Dict, List, Set

from pydantic.fields import Undefined

from pait import field
from pait.model.status import PaitStatus
from pait.util import gen_example_json_from_schema

from ..model.core import PaitCoreModel
from .base_parse import PaitBaseParse  # type: ignore


class PaitMd(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: str = "Pait Doc",
        use_html_details: bool = True,
    ):
        self._use_html_details: bool = use_html_details  # some not support markdown in html
        self._title: str = title

        super().__init__(pait_dict)

        self._field_name_set: Set[str] = set()
        for field_class_name in dir(field):
            class_: type = getattr(field, field_class_name, None)
            if inspect.isclass(class_) and issubclass(class_, field.BaseField) and class_ != field.BaseField:
                self._field_name_set.add(field_class_name.lower())

        self.content = self.gen_markdown_text()
        self._content_type = ".md"

    @staticmethod
    def gen_md_param_table(field_dict_list: List[dict], blank_num: int = 8) -> str:
        """
        gen param md table
        :param field_dict_list:
        :param blank_num: table str indent
        :return:
            |param name|type|default value|description|other|
            |---|---|---|---|---|
            |age|string|10|user age||
        """
        blank_num_str: str = " " * blank_num
        markdown_text: str = f"{blank_num_str}|param name|type|default value|description|other|\n"
        markdown_text += f"{blank_num_str}|---|---|---|---|---|\n"
        field_dict_list = sorted(field_dict_list, key=lambda x: x["param_name"])
        for field_info_dict in field_dict_list:
            default = field_info_dict["default"]
            if default is Undefined:
                default = "**`Required`**"
            type_ = field_info_dict["type"]
            if default is Undefined:
                type_ = "**`Required`**"
            markdown_text += (
                f"{blank_num_str}"
                f"|{field_info_dict['param_name']}"
                f"|{type_}"
                f"|{default}"
                f"|{field_info_dict['description']}"
                f"|{field_info_dict['other'] or ''}"
                f"|\n"
            )
        return markdown_text

    def gen_markdown_text(self) -> str:
        markdown_text: str = f"# {self._title}\n"
        for group in self._group_list:
            # group
            if self._use_html_details:
                markdown_text += f"<details><summary>Group: {group}</summary>\n\n"
            else:
                markdown_text += f"## Group: {group}\n\n"
            for pait_model in self._group_pait_dict[group]:
                # func info
                markdown_text += f"### Name: {pait_model.operation_id}\n\n"
                status_text: str = ""
                if pait_model.status in (PaitStatus.test, PaitStatus.design, PaitStatus.dev, PaitStatus.integration):
                    status_text = f"<font color=#00BFFF>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.release, PaitStatus.complete):
                    status_text = f"<font color=#32CD32>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.abandoned, PaitStatus.abnormal):
                    status_text = f"<font color=#DC143C>{pait_model.status.value}</font>"
                elif pait_model.status:
                    status_text = f"{pait_model.status.value}"

                if pait_model.desc:
                    markdown_text += f"\n\n**Desc**:{pait_model.desc}\n\n"

                # func or interface details
                func_code: CodeType = pait_model.func.__code__
                markdown_text += "|Author|Status|func|summary|\n"
                markdown_text += "|---|---|---|---|\n"
                markdown_text += (
                    f"|{','.join(pait_model.author)}"
                    f"|{status_text}"
                    f'|<abbr title="file:{pait_model.func_path or func_code.co_filename};'
                    f'line: {func_code.co_firstlineno}">'
                    f"{pait_model.func.__qualname__}</abbr>"
                    f"|{pait_model.summary}|\n"
                )

                # request info
                markdown_text += f"- Path: {pait_model.path}\n"
                markdown_text += f"- Method: {','.join(pait_model.method_list)}\n"
                markdown_text += "- Request:\n"

                field_dict: Dict[str, List[Dict[str, Any]]] = self._parse_func_param_to_field_dict(pait_model.func)
                # request body info
                field_key_list: List[str] = sorted(field_dict.keys())
                for _field in field_key_list:
                    if _field.lower() not in self._field_name_set:
                        continue
                    field_dict_list = field_dict[_field]
                    markdown_text += f"{' ' * 4}- {_field.capitalize()} Param\n\n"
                    markdown_text += self.gen_md_param_table(field_dict_list)

                # response info
                markdown_text += "- Response:\n\n"
                if pait_model.response_model_list:
                    for resp_model_class in pait_model.response_model_list:
                        resp_model = resp_model_class()
                        markdown_text += f"{' ' * 4}- {resp_model.name or resp_model.__class__.__name__}\n\n"
                        markdown_text += f"{' ' * 8}|status code|media type|description|\n"
                        markdown_text += f"{' ' * 8}|---|---|---|\n"
                        markdown_text += (
                            f"{' ' * 8}|{','.join([str(i) for i in resp_model.status_code])}"
                            f"|{resp_model.media_type}"
                            f"|{resp_model.description}"
                            f"|\n"
                        )
                        if resp_model.header:
                            markdown_text += f"{' ' * 8}- Header\n"
                            markdown_text += f"{' ' * 12}{resp_model.header}\n"
                        if resp_model.response_data:
                            markdown_text += f"{' ' * 8}- Response Data\n\n"
                            field_dict_list = self._parse_schema(resp_model.response_data.schema())
                            markdown_text += self.gen_md_param_table(field_dict_list, blank_num=12)
                            markdown_text += f"{' ' * 8}- Example Response Json Data\n\n"

                            example_dict = gen_example_json_from_schema(resp_model.response_data.schema(), False)
                            blank_num_str: str = " " * 12
                            json_str: str = "\n".join(
                                [blank_num_str + i for i in json.dumps(example_dict, indent=2).split("\n")]
                            )
                            markdown_text += f"{blank_num_str}```json\n{json_str}\n{blank_num_str}```\n\n"
                markdown_text += "\n"
            if self._use_html_details:
                markdown_text += "</details>"
        return markdown_text



import inspect
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pait.app.base import BaseAppHelper
from pait.exceptions import CheckValueError
from pait.g import pait_data
from pait.model.core import PaitCoreModel
from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus
from pait.param_handle import async_class_param_handle, async_func_param_handle, class_param_handle, func_param_handle
from pait.util import FuncSig, gen_tip_exc, get_func_sig


def _check_at_most_one_of(at_most_one_of_list: List[List[str]], func_kwargs: Dict[str, Any]) -> None:
    """Check whether each group of parameters appear at the same time"""
    for at_most_one_of in at_most_one_of_list:
        if len([i for i in at_most_one_of if func_kwargs.get(i, None) is not None]) > 1:
            raise CheckValueError(f"requires at most one of param {' or '.join(at_most_one_of)}")


def _check_required_by(required_by: Dict[str, List[str]], func_kwargs: Dict[str, Any]) -> None:
    """Check dependencies between parameters"""
    for pre_param, param_list in required_by.items():
        if pre_param not in func_kwargs:
            continue
        for param in param_list:
            if func_kwargs.get(param, None) is not None and func_kwargs[pre_param] is None:
                raise CheckValueError(f"{pre_param} requires param {' and '.join(param_list)}, which if not none")


def pait(
    app_helper_class: "Type[BaseAppHelper]",
    make_mock_response_fn: Callable[[Type[PaitResponseModel]], Any],
    # param check
    at_most_one_of_list: Optional[List[List[str]]] = None,
    required_by: Optional[Dict[str, List[str]]] = None,
    # doc
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    summary: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
) -> Callable:
    if not isinstance(app_helper_class, type):
        raise TypeError(f"{app_helper_class} must be class")
    if not issubclass(app_helper_class, BaseAppHelper):
        raise TypeError(f"{app_helper_class} must sub from {BaseAppHelper.__class__.__name__}")
    app_name: str = app_helper_class.app_name

    def wrapper(func: Callable) -> Callable:
        func_sig: FuncSig = get_func_sig(func)

        pait_core_model: PaitCoreModel = PaitCoreModel(
            author=author,
            app_helper_class=app_helper_class,
            make_mock_response_fn=make_mock_response_fn,
            desc=desc,
            summary=summary,
            func=func,
            func_name=name,
            status=status,
            group=group,
            tag=tag,
            response_model_list=response_model_list,
        )
        pait_data.register(app_name, pait_core_model)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                try:
                    # only use in runtime, support cbv
                    class_ = getattr(inspect.getmodule(func), pait_core_model.qualname)
                    # real param handle
                    app_helper: BaseAppHelper = app_helper_class(class_, args, kwargs)  # type: ignore
                    # auto gen param from request
                    func_args, func_kwargs = await async_func_param_handle(app_helper, func_sig)
                    # support sbv
                    await async_class_param_handle(app_helper)
                    # param check
                    if at_most_one_of_list:
                        _check_at_most_one_of(at_most_one_of_list, func_kwargs)
                    if required_by:
                        _check_required_by(required_by, func_kwargs)
                except Exception as e:
                    raise e from gen_tip_exc(func, e)
                return await pait_core_model.func(*func_args, **func_kwargs)

            return dispatch
        else:

            @wraps(func)
            def dispatch(*args: Any, **kwargs: Any) -> Callable:
                try:
                    # only use in runtime
                    class_ = getattr(inspect.getmodule(func), pait_core_model.qualname)
                    # real param handle
                    app_helper = app_helper_class(class_, args, kwargs)  # type: ignore
                    # auto gen param from request
                    func_args, func_kwargs = func_param_handle(app_helper, func_sig)
                    # support sbv
                    class_param_handle(app_helper)
                    # param check
                    if at_most_one_of_list:
                        _check_at_most_one_of(at_most_one_of_list, func_kwargs)
                    if required_by:
                        _check_required_by(required_by, func_kwargs)
                except Exception as e:
                    raise e from gen_tip_exc(func, e)
                return pait_core_model.func(*func_args, **func_kwargs)

            return dispatch

    return wrapper



from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from json import JSONEncoder
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type

from pait.model.response import PaitResponseModel
from pait.model.status import PaitStatus


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if type(obj) == date:
            return datetime.strptime(str(obj), "%Y-%m-%d")
        elif isinstance(obj, datetime):
            # 返回时间戳格式
            return int(obj.timestamp())
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, Enum):
            return obj.value
        else:
            return super().default(obj)


class Config(object):
    """
    Provide pait parameter configuration, the init_config method is valid only before the application service is started
    """

    __initialized: bool = False

    def __init__(self) -> None:
        self.author: Tuple[str, ...] = ("",)
        self.status: PaitStatus = PaitStatus.undefined
        self.enable_mock_response: bool = False
        self.enable_mock_response_filter_fn: Optional[Callable[[Type[PaitResponseModel]], bool]] = None
        self.block_http_method_set: Set[str] = set()
        self.default_response_model_list: List[Type[PaitResponseModel]] = []
        self.json_encoder: Type[JSONEncoder] = CustomJSONEncoder
        self.json_type_default_value_dict: Dict[str, Any] = {
            "null": None,
            "bool": True,
            "boolean": True,
            "string": "",
            "number": 0.0,
            "float": 0.0,
            "integer": 0,
            "object": {},
            "array": [],
        }
        self.python_type_default_value_dict: Dict[type, Any] = {
            bool: True,
            float: 0.0,
            int: 0,
            str: "",
            list: [],
            tuple: [],
            datetime: 0,
            Decimal: 0,
        }

    def __setattr__(self, key: Any, value: Any) -> None:
        if not self.__initialized:
            super().__setattr__(key, value)
        else:
            raise RuntimeError("Can not set new value in runtime")

    def init_config(
        self,
        author: Tuple[str, ...] = ("",),
        status: PaitStatus = PaitStatus.undefined,
        block_http_method_set: Optional[Set[str]] = None,
        default_response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
        json_type_default_value_dict: Optional[Dict[str, Any]] = None,
        python_type_default_value_dict: Optional[Dict[type, Any]] = None,
        json_encoder: Type[JSONEncoder] = CustomJSONEncoder,
        enable_mock_response: bool = False,
        enable_mock_response_filter_fn: Optional[Callable[[Type[PaitResponseModel]], bool]] = None,
    ) -> None:
        """
        :param author:  Only @pait(author=None) will be called to change the configuration
        :param status:  Only @pait(status=None) will be called to change the configuration
        :param block_http_method_set:
            Under normal circumstances, pait.load_app can obtain the http method of the routing handle.
            However, some application frameworks such as flask will automatically add optional http methods
             to the handle, which is great, but you may not want to use them in pait, and pait will not
             automatically recognize them, so you can use parameters to disable certain http method
        :param default_response_model_list: Add a default response structure for all routing handles
        :param json_type_default_value_dict: Configure default values for each json type
        :param python_type_default_value_dict: Configure default values for each python type
        :param json_encoder: Define certain types of serialization rules
        :param enable_mock_response:
            If True, the routing handle will return an example value, which is automatically generated by
             the first response body in response_list
        :param enable_mock_response_filter_fn:
            There are multiple response bodies in response_list, you can filter by changing the function
        :return:
        """
        self.author = author
        self.status = status
        self.block_http_method_set = block_http_method_set or set()
        self.default_response_model_list = default_response_model_list or []
        if json_type_default_value_dict:
            self.json_type_default_value_dict.update(json_type_default_value_dict)
        if python_type_default_value_dict:
            self.python_type_default_value_dict.update(python_type_default_value_dict)
        self.json_encoder = json_encoder
        self.enable_mock_response = enable_mock_response
        self.enable_mock_response_filter_fn = enable_mock_response_filter_fn

        self.__initialized = True

    @property
    def initialized(self) -> bool:
        """Judge whether it has been initialized, it can only be initialized once per run"""
        return self.__initialized



class PaitBaseException(Exception):
    pass


class NotFoundFieldError(PaitBaseException):
    pass


class NotFoundValueError(PaitBaseException):
    pass


class CheckValueError(PaitBaseException):
    pass
